"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.mockServer = mockServer;

var _graphqlTools = require("graphql-tools");

var _graphql = require("graphql");

var _automock = require("./automock");

var _list = require("./list");

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function mockServer(schemaDefinition, mocks = {}, automocks = _automock.defaultAutomocks) {
  const schema = (0, _graphqlTools.buildSchemaFromTypeDefinitions)(schemaDefinition);

  if (automocks) {
    addAutomocks(schema, mocks, automocks);
  }

  validateMocks(mocks, schema);
  forEachField(schema, (type, field) => {
    field.resolve = getFieldResolver(type, field, mocks);
  });
  forEachInterface(schema, interfaceType => {
    interfaceType.resolveType = interfaceResolveType;
  });
  return (query, variables = {}, mockOverride = {}) => {
    const result = (0, _graphql.graphqlSync)(schema, query, // TODO Rename to mockOverride
    {
      queryMock: mockOverride
    }, {}, variables);
    throwUnexpectedErrors(result);
    return result;
  };
}

function addAutomocks(schema, mocks, getMocks) {
  forEachField(schema, (parentType, field) => {
    if (mocks[parentType.name] && mocks[parentType.name][field.name]) {
      return;
    }

    for (let getMock of getMocks) {
      const baseMock = getMock(parentType, field, getMocks);

      if (baseMock) {
        mocks[parentType.name] = mocks[parentType.name] || {};
        mocks[parentType.name][field.name] = baseMock;
        return;
      }
    }
  });
}
/**
 * In order to resolve interface field, we curretly require users
 * to specify the type to resolve to via `__typename` in `queryMock`.
 *
 * Future work:
 * We may want to remove that requirement so queries can always resolve
 * succesfully without having to specify a `queryMock`.
 * Here are some ideas:
 * - Have this function pick randomly a valid type
 * - Have this function look at the query fragment and pick the most likely type
 * - Have this function look at the query fragment, pick its type if there is only one,
 *   otherwise throw a helpful validation error.
 * - Have this function be customizable via MockMap / baseMock
 * - Have this function be customizable via autoMock
 */


const interfaceResolveType = markUnexpectedErrors(source => {
  if (!source.queryMock || !source.queryMock.__typename) {
    throw Error('queryMock must specify type for interface fields.');
  }

  return source.queryMock.__typename;
});

function getFieldResolver(type, field, mocks) {
  return markUnexpectedErrors((source, args, context, info) => {
    const baseMock = getFieldMock(type, field, mocks);
    const parentMock = source.parentMock && source.parentMock[field.name];
    const mergedBaseMocks = mergeBaseMocks(baseMock, parentMock, field.type, {
      parentType: type,
      field: field,
      path: undefined
    });
    const mergedBaseMocksValue = mergedBaseMocks(args); // TODO check this

    const fieldName = typeof info.path.key === 'string' ? info.path.key : field.name;
    const queryMockValue = source.queryMock ? getMockValue(source.queryMock[fieldName], args) : undefined;

    if (!((0, _graphql.getNullableType)(field.type) instanceof _graphql.GraphQLList)) {
      // TODO remove args?
      return mergeMockValues(mergedBaseMocksValue, queryMockValue, field);
    }

    if (queryMockValue === null || queryMockValue instanceof Error) {
      return queryMockValue;
    }

    const mergedArray = [];
    let queryMockList = queryMockValue;

    if (Array.isArray(queryMockList)) {
      const queryMockArray = queryMockList;
      queryMockList = new _list.MockList(queryMockArray.length, ({}, index) => queryMockArray[index]);
    }

    if (queryMockValue === undefined) {
      queryMockList = new _list.MockList(mergedBaseMocksValue.length, () => {});
    }

    for (let index = 0; index < queryMockList.length; index++) {
      const mockListItemValue = mergedBaseMocksValue.mockFunction(args, index);
      const queryMockListItemValue = queryMockList.mockFunction(args, index);
      mergedArray.push(mergeMockValues(mockListItemValue, queryMockListItemValue, field));
    }

    return mergedArray;
  });
}

function mergeMockValues(baseMockValue, queryMockValue, field) {
  if (queryMockValue === null) {
    return null;
  }

  if (queryMockValue === undefined && baseMockValue instanceof Error) {
    return baseMockValue;
  }

  if ((0, _graphql.isLeafType)((0, _graphql.getNamedType)(field.type))) {
    if (queryMockValue === undefined) {
      return baseMockValue;
    }

    return queryMockValue;
  }

  return {
    queryMock: queryMockValue,
    parentMock: baseMockValue
  };
}

function mergeBaseMocks(baseMock, overrideMock, graphQLType, baseMockInfo) {
  return function (...args) {
    let baseMockValue = getBaseMockValue(graphQLType, baseMockInfo, baseMock, ...args);
    let overrideMockValue = getMockValue(overrideMock, ...args);
    const nullableType = (0, _graphql.getNullableType)(graphQLType);

    if (overrideMockValue === null || overrideMockValue instanceof Error || (0, _graphql.isLeafType)(nullableType) && overrideMockValue !== undefined) {
      return overrideMockValue;
    }

    if (overrideMockValue === undefined && (0, _graphql.isLeafType)(nullableType)) {
      return baseMockValue;
    }

    if (overrideMockValue === undefined && baseMockValue instanceof Error) {
      return baseMockValue;
    }

    if (baseMockValue === undefined) {
      return overrideMockValue;
    }

    if (nullableType instanceof _graphql.GraphQLList) {
      if (overrideMockValue === undefined) {
        overrideMockValue = new _list.MockList(baseMockValue.length);
      }

      const baseMockList = baseMockValue;
      let overrideMockList = overrideMockValue;
      const overrideMockListFunction = overrideMockList.mockFunction;
      const baseMockListFunction = baseMockList.mockFunction;
      return new _list.MockList(overrideMockList.length, (args, index) => {
        const graphQLItemType = (0, _graphql.getNamedType)(graphQLType);
        return mergeBaseMocks(baseMockListFunction, overrideMockListFunction, graphQLItemType, _objectSpread({}, baseMockInfo, {
          path: {
            prev: baseMockInfo.path,
            key: `${index}`
          }
        }))(args, index);
      });
    } // (nullableType instanceof GraphQLObjectType) === true


    if (overrideMockValue === undefined) {
      overrideMockValue = {};
    }

    const mergedMockObject = {};
    const mergedObjectObjectKeys = new Set(Object.keys(overrideMockValue).concat(Object.keys(baseMockValue)));
    const overrideMockValueCopy = overrideMockValue;
    const baseMockValueCopy = baseMockValue;
    mergedObjectObjectKeys.forEach(nestedFieldName => {
      const nestedField = nullableType.getFields()[nestedFieldName];

      mergedMockObject[nestedFieldName] = (...mergedMockArgs) => {
        return mergeBaseMocks(baseMockValueCopy[nestedFieldName], overrideMockValueCopy[nestedFieldName], nestedField.type, _objectSpread({}, baseMockInfo, {
          path: {
            prev: baseMockInfo.path,
            key: nestedFieldName
          }
        }))(...mergedMockArgs);
      };
    });
    return mergedMockObject;
  };
} // TODO Rename to getMockValue
// TODO Add path to the mock. Currently, we only show the Parent name,
// the field name, and the path inside the merged


function getBaseMockValue(graphQLType, baseMockInfo, baseMock, ...baseMockArgs) {
  const nullableType = (0, _graphql.getNullableType)(graphQLType);
  const path = baseMockInfo.path;
  let baseMockValue;

  try {
    baseMockValue = getMockValue(baseMock, ...baseMockArgs);
  } catch (err) {
    throw Error(`Base mock for \'${getFieldName(baseMockInfo)}\' ` + `threw an error for path '${getFullPath(path)}'.\n` + `Base mocks are not allowed to throw errors. ` + `In the rare case you actually want a base mock to return a GraphQL error, ` + `have the base mock return an Error() instead of throwing one.\n` + `Original error:\n${err}`);
  }

  if (baseMockValue === undefined && (nullableType instanceof _graphql.GraphQLObjectType || nullableType instanceof _graphql.GraphQLInterfaceType)) {
    return baseMockValue;
  }

  if (baseMockValue instanceof Error) {
    return baseMockValue;
  }

  if (baseMockValue instanceof Promise) {
    throw Error(`Base mock for \'${getFieldName(baseMockInfo)}\' ` + `returned a promise for path '${getFullPath(path)}'.\n` + `Mock functions must be synchronous.`);
  }

  if (baseMockValue === undefined) {
    if (baseMockInfo.path) {
      return baseMockValue;
    } else {
      throw Error(`Base mock for \'${getFieldName(baseMockInfo)}\' ` + `returned 'undefined'.\n` + `Base mocks are not allowed to return 'undefined'. ` + `Return a value compatible with type '${nullableType.name}'.`);
    }
  }

  if (baseMockValue === null) {
    throw Error(`Base mock for \'${getFieldName(baseMockInfo)}\' ` + `returned 'null' for path '${getFullPath(path)}'.\n` + `Base mocks are not allowed to return 'null'. ` + `Use 'mockOverride' to specify 'null' values instead.`);
  }

  if ((0, _graphql.isLeafType)(nullableType)) {
    try {
      nullableType.serialize(baseMockValue);
    } catch (err) {
      throw Error(`Base mock for \'${getFieldName(baseMockInfo)}\' ` + `returned an invalid value for path '${getFullPath(path)}'.\n` + `Value '${baseMockValue}' is incompatible with type '${nullableType.name}'.`);
    }
  }

  if (nullableType instanceof _graphql.GraphQLObjectType) {
    if (typeof baseMockValue !== 'object') {
      throw Error(`Base mock for \'${getFieldName(baseMockInfo)}\' ` + `did not return an object for path '${getFullPath(path)}'.\n` + `Value '${baseMockValue}' is incompatible with type '${nullableType.name}'.`);
    }

    const fieldNames = nullableType.getFields();
    Object.keys(baseMockValue).forEach(key => {
      if (!fieldNames[key]) {
        const nestedFullPath = getFullPath({
          key,
          prev: baseMockInfo.path
        });
        throw Error(`Base mock for \'${getFieldName(baseMockInfo)}\' ` + `returns a value for field path \'${nestedFullPath}\' that does not exist. ` + 'Base mocks should return values only for valid fields.');
      }
    });
  }

  if (nullableType instanceof _graphql.GraphQLList && !(baseMockValue instanceof _list.MockList) && !baseMockInfo.path) {
    throw Error(`Base mock for \'${getFieldName(baseMockInfo)}\' ` + `did not return a MockList for path '${getFullPath(path)}'.\n` + `Use 'mockList' function to mock lists in base mocks.`);
  }

  return baseMockValue;
}

function getFullPath(fieldPath) {
  const allPaths = [];
  let currentPath = fieldPath;

  while (currentPath) {
    allPaths.push(currentPath.key);
    currentPath = currentPath.prev;
  }

  allPaths.reverse();
  return allPaths.join('.');
}

function getFieldName(baseMockInfo) {
  return `${baseMockInfo.parentType.name}.${baseMockInfo.field.name}`;
}

function getMockValue(mock, ...args) {
  return typeof mock === 'function' ? mock(...args) : mock;
}

function getFieldMock(type, field, mocks) {
  let baseMock;

  if (mocks[type.name]) {
    baseMock = mocks[type.name][field.name];
  }

  if (baseMock) {
    return baseMock;
  }

  const fieldInterfaces = [];
  type.getInterfaces().forEach(parentInterface => {
    if (parentInterface.getFields()[field.name]) {
      fieldInterfaces.push(parentInterface.name);
    }
  });

  if (fieldInterfaces.length > 1) {
    throw Error('More than 1 interface for this field. Define base mock on the type.');
  }

  if (fieldInterfaces.length === 1) {
    if (mocks[fieldInterfaces[0]] && mocks[fieldInterfaces[0]][field.name]) {
      return mocks[fieldInterfaces[0]][field.name];
    }
  }

  if ((0, _graphql.getNullableType)(field.type) instanceof _graphql.GraphQLList) {
    throw Error(`There is no base mock for '${type.name}.${field.name}'. ` + `All queried list fields must have a base mock defined using mockList.`);
  }

  if ((0, _graphql.isLeafType)((0, _graphql.getNamedType)(field.type))) {
    throw Error(`There is no base mock for '${type.name}.${field.name}'. ` + `All queried fields must have a base mock.`);
  }
} // Input Validation


function validateMocks(mocks, schema) {
  const typeMap = schema.getTypeMap();
  Object.keys(mocks).forEach(typeName => {
    if (!typeMap[typeName]) {
      throw Error(`mocks['${typeName}'] is not defined in schema.`);
    }

    if (!(typeMap[typeName] instanceof _graphql.GraphQLInterfaceType || typeMap[typeName] instanceof _graphql.GraphQLObjectType)) {
      throw Error('baseMock can only define field mocks on Type or Interface.');
    }

    if (typeof mocks[typeName] !== 'object') {
      throw Error('mocks should be an object of object of functions.');
    }

    const isInterface = typeMap[typeName] instanceof _graphql.GraphQLInterfaceType;
    const fields = typeMap[typeName].getFields();
    Object.keys(mocks[typeName]).forEach(fieldName => {
      if (!fields[fieldName]) {
        throw Error(`mocks['${typeName}']['${fieldName}'] is not defined in schema.`);
      }

      if (typeof mocks[typeName][fieldName] !== 'function') {
        // TODO Add better validation message
        throw Error('mocks should be an object of object of functions.');
      }

      if (isInterface && !(0, _graphql.isLeafType)((0, _graphql.getNullableType)(fields[fieldName].type))) {
        // TODO Add better validation message
        throw Error('It is not allowed to define mocks for non-leaf fields on interfaces.');
      }
    });
  });
} // Error Utils


function markUnexpectedErrors(fieldResolver) {
  return function (...args) {
    try {
      return fieldResolver(...args);
    } catch (error) {
      error.throwError = true;
      throw error;
    }
  };
}

function throwUnexpectedErrors(result) {
  if (!result.errors) {
    return;
  }

  result.errors.forEach(error => {
    if (error.originalError && error.originalError.throwError) {
      throw error.originalError;
    }
  });
} // Schema utils


function forEachField(schema, callback) {
  const typeMap = schema.getTypeMap();
  Object.keys(typeMap).forEach(typeName => {
    const type = typeMap[typeName];

    if ((0, _graphql.getNamedType)(type).name.startsWith('__') || !(type instanceof _graphql.GraphQLObjectType)) {
      return;
    }

    const fields = type.getFields();
    Object.keys(fields).forEach(fieldName => {
      const field = fields[fieldName];
      callback(type, field);
    });
  });
}

function forEachInterface(schema, callback) {
  const typeMap = schema.getTypeMap();
  Object.keys(typeMap).forEach(typeName => {
    const type = typeMap[typeName];

    if ((0, _graphql.getNamedType)(type).name.startsWith('__') || !(type instanceof _graphql.GraphQLInterfaceType)) {
      return;
    }

    callback(type);
  });
}
//# sourceMappingURL=server.js.map